From 6a5dd5766a5806258d665cab343a103b5705bf32 Mon Sep 17 00:00:00 2001
From: Ian Jackson <ijackson@chiark.greenend.org.uk>
Date: Fri, 21 Jan 2022 19:42:43 +0000
Subject: [PATCH 2/4] * inotify.c: Use a buffer from the heap

We are going to need this buffer to be persistent, because we may need
to stop processing it halfway through (if the main input event buffer
fills up).

Also, getting a value from FIONREAD and then feeding it to alloca is
even less wise than feeding it to malloc.
---
 src/inotify.c | 17 +++++++++++------
 1 file changed, 11 insertions(+), 6 deletions(-)

diff --git a/src/inotify.c b/src/inotify.c
index b676ea216a..ad02164b37 100644
--- a/src/inotify.c
+++ b/src/inotify.c
@@ -43,6 +43,9 @@ Copyright (C) 2012-2022 Free Software Foundation, Inc.
 /* File handle for inotify.  */
 static int inotifyfd = -1;
 
+static size_t inotify_buffer_len = 0;
+static char *inotify_buffer = NULL;
+
 /* Alist of files being watched.  We want the returned descriptor to
    be unique for every watch, but inotify returns the same descriptor
    WD for multiple calls to inotify_add_watch with the same file.
@@ -319,9 +322,13 @@ inotify_callback (int fd, void *_)
   // FIXME: feeding the value from FIONREAD to a memory allocator is unwise
   // Sadly the inotify event buffer parsing algorithm is not capable of
   // handling the case where we read only part of a kernel event.
-  USE_SAFE_ALLOCA;
-  char *buffer = SAFE_ALLOCA (to_read);
-  ssize_t n = read (fd, buffer, to_read);
+  if (to_read > inotify_buffer_len)
+    {
+      inotify_buffer = xrealloc(inotify_buffer, to_read);
+      inotify_buffer_len = to_read;
+    }
+
+  ssize_t n = read (fd, inotify_buffer, to_read);
   if (n < 0)
     report_file_notify_error ("Error while reading file system events", Qnil);
 
@@ -331,7 +338,7 @@ inotify_callback (int fd, void *_)
 
   for (ssize_t i = 0; i < n; )
     {
-      struct inotify_event *ev = (struct inotify_event *) &buffer[i];
+      struct inotify_event *ev = (struct inotify_event *) &inotify_buffer[i];
       Lisp_Object descriptor = INT_TO_INTEGER (ev->wd);
       Lisp_Object prevtail = find_descriptor (descriptor);
 
@@ -351,8 +358,6 @@ inotify_callback (int fd, void *_)
         }
       i += sizeof (*ev) + ev->len;
     }
-
-  SAFE_FREE ();
 }
 
 DEFUN ("inotify-add-watch", Finotify_add_watch, Sinotify_add_watch, 3, 3, 0,
-- 
2.20.1

